"""
Command line tool.
"""
import sys
from dox.config import init_environment, get_keyfields, get_cfg, write_keyfields, get_keymap, write_keymap, clean_hashes, is_modified, write_hash
from dox.client import ping_library, get_content_keys, get_content_item
from os import walk, getcwd
from dox.uploader import upload_document, find_key, extract_keyfield
import os.path

def init(args):
    """
    Initialize the Dox environment.
    """
    if not (args.library_key and args.project and args.content_type and args.body_field):
        print 'You must specify the library key, project, content type and body field to initialize the environment.'
        sys.exit(1)
    
    print 'Dox -----------------------------------------'
    print 'Initializing environment with:'
    print 'Library Key',args.library_key
    print 'Project',args.project
    print 'Content Type',args.content_type
    print 'Body Field',args.body_field
    print 'Key Field',args.key_field
    print '---------------------------------------------'
    
    init_environment(args)
    
    print 'Environment initialized.'
    
    print 'Testing environment...'
    ping_library()
    print 'Connection settings are good.'

def up(args):
    """
    Alias for upload.
    """
    return upload(args)

def upload(args):
    """
    Uploads documents.
    """
    print 'Uploading documents...'
    keymap = get_keymap()
    keyfields = get_keyfields()
    for root, dirs, files in walk(getcwd()):
        if 'env' in dirs:
            dirs.remove('env') # don't walk into environment
        
        for name in files:
            if name.endswith('.md'):
                path = os.path.join(root,name)
                if is_modified(path):
                    print 'Uploading',name
                    key = find_key(path,keymap,keyfields)
                    key, created = upload_document(path,key=key)
                    write_hash(path)
                    if created:
                        print 'Created new content item',key
                        keymap[path] = key
                        keyfield = extract_keyfield(path)
                        print 'assigning key',key,'to keyfields',keyfields,'under keyfield',keyfield
                        keyfields[keyfield] = key
                else:
                    print name, 'not modified. Skipping.'
    
    write_keyfields(keyfields)
    write_keymap(keymap)


def keyfields(args):
    """
    Synchronizes the local cache of keyfield / key data.
    """
    print 'Synchronizing keyfield cache.'
    cfg = get_cfg()
    keyfield_name = cfg.get('Connection','key_field')
    keyfield_data = {}
    keys = get_content_keys()
    for key in keys:
        content_item = get_content_item(key)
        keyfield_data[content_item['data'][keyfield_name]] = key
        print 'Mapping',content_item['data'][keyfield_name],'to',key
    write_keyfields(keyfield_data)
    print 'Keyfield cache synchronized.'

def clean(args):
    """
    Cleans out the local hash directory - essentially marking all
    the local markdown files as modified.
    """
    print 'Cleaning out local file records.  All local files eligible for upload.'
    clean_hashes()



"""
Config for Dox.
"""
from ConfigParser import SafeConfigParser
import os.path
from os import getcwd, mkdir, remove, walk
import json
import hashlib

def dox_dir():
    """
    Gets or creates the .dox directory.
    """
    dox_dirpath = os.path.join(getcwd(),'.dox')
    
    if not os.path.exists(dox_dirpath):
        mkdir(dox_dirpath)
    
    return dox_dirpath

def init_environment(args):
    """
    Initializes the environment.
    """    
    dirpath = dox_dir()
    
    cfg_path = os.path.join(dirpath,'dox.cfg')
    
    # Clean out previous config
    if os.path.exists(cfg_path):
        remove(cfg_path)
    
    # New Config File
    cfg = SafeConfigParser()
    cfg.add_section('Connection')
    cfg.set('Connection','endpoint',args.endpoint)
    cfg.set('Connection','library_key',args.library_key)
    cfg.set('Connection','project',args.project)
    cfg.set('Connection','content_type',args.content_type)
    cfg.set('Connection','body_field',args.body_field)
    if args.key_field:
        cfg.set('Connection','key_field',args.key_field)
    
    if args.api_key:
        cfg.set('Connection','api_key',args.api_key)
    
    with open(cfg_path,'wb') as cfg_file:
        cfg.write(cfg_file)
        
    # Make file hash directory
    hashdir_path = os.path.join(dirpath,'hashes')
    if not os.path.exists(hashdir_path):
        mkdir(hashdir_path)

def get_cfg():
    """
    Gets the config file.
    """
    cfg = SafeConfigParser()
    cfg_path = os.path.join(dox_dir(),'dox.cfg')

    # Sanity check
    if not os.path.exists(cfg_path):
        raise ValueError('No config file found.  Directory has not been initialized for dox.  Use dox init.')

    cfg.read(cfg_path)
    
    return cfg

def get_keyfields():
    """
    Gets the keyfields data.
    """
    dirpath = dox_dir()
    keyfield_path = os.path.join(dirpath,'keyfields.json')
    if os.path.exists(keyfield_path):
        with open(keyfield_path,'r') as keyfield_file:
            keyfield_data = json.loads(keyfield_file.read())
            return keyfield_data
    else:
        return {}

def write_keyfields(data):
    """
    Writes the keyfield data file.
    """
    dirpath = dox_dir()
    keyfield_path = os.path.join(dirpath,'keyfields.json')
    with open(keyfield_path,'wb') as keyfield_file:
        keyfield_file.write(json.dumps(data))

def get_keymap():
    """
    Gets the keymap data.
    """
    dirpath = dox_dir()
    keymap_path = os.path.join(dirpath,'keymap.json')
    if os.path.exists(keymap_path):
        with open(keymap_path,'r') as keymap_file:
            keymap_data = json.loads(keymap_file.read())
            return keymap_data
    else:
        return {}

def write_keymap(data):
    """
    Saves the keymap data.
    """
    dirpath = dox_dir()
    keymap_path = os.path.join(dirpath,'keymap.json')
    with open(keymap_path,'wb') as keymap_file:
        keymap_file.write(json.dumps(data))


def clean_hashes():
    """
    Cleans the local file hash directory out.
    """
    hash_path = os.path.join(dox_dir(),'hashes')
    if os.path.exists(hash_path):
        for root, dirs, files in walk(hash_path):
            for name in files:
                if name.endswith('.hash'):
                    remove(os.path.join(root,name))
    else:
        mkdir(hash_path)

def write_hash(markdown_file_path):
    """
    Scans the file and records a hash digest of the contents.
    """
    with open(markdown_file_path) as markdown_file:
        d = hashlib.sha256()
        d.update(markdown_file.read())
        digest = d.hexdigest()
        hash_file_path = '%s.hash' % os.path.join(dox_dir(),'hashes',os.path.split(markdown_file.name)[1])
        with open(hash_file_path,'wb') as hash_file:
            hash_file.write(digest)

def is_modified(markdown_file_path):
    """
    Tests if the markdown file has been modified.
    """
    with open(markdown_file_path,'r') as markdown_file:
        hashfile_path = '%s.hash' % os.path.join(dox_dir(),'hashes',os.path.split(markdown_file.name)[1])
        if os.path.exists(hashfile_path):
            d = hashlib.sha256()
            d.update(markdown_file.read())
            digest = d.hexdigest()
            with open(hashfile_path) as hashfile:
                stored_hash = hashfile.read()
                if stored_hash != digest:
                    return True # non-matching hashes - file is modified
                else:
                    return False # hashes match - file has not been modified
        else:
            return True # no stored hash - file is modified by definition



"""
Axilent Client functionality for Dox.
"""
from sharrock.client import HttpClient, ResourceClient, ServiceException
from dox.config import get_cfg
from django.template.defaultfilters import slugify

def _get_resource(app,resource,library=True):
    """
    Gets a resource client.
    """
    cfg = get_cfg()
    apikey_setting = 'library_key' if library else 'api_key'
    return ResourceClient('%s/api/resource' % cfg.get('Connection','endpoint'),app,'beta3',resource,auth_user=cfg.get('Connection',apikey_setting))

def _get_client(app,library=True):
    """
    Gets a regular API client.
    """
    cfg = get_cfg()
    apikey_setting = 'library_key' if library else 'api_key'
    return HttpClient('%s/api' % cfg.get('Connection','endpoint'),app,'beta3',auth_user=cfg.get('Connection',apikey_setting))

def get_content_library_resource():
    """
    Gets a content library resource.
    """
    return _get_resource('axilent.library','content')

def get_library_client():
    """
    Gets the library API client.
    """
    return _get_client('axilent.library')

def ping_library():
    """
    Pings the library.
    """
    cfg = get_cfg()
    lib = get_library_client()
    lib.ping(project=cfg.get('Connection','project'),content_type=cfg.get('Connection','content_type'))

def get_content_api():
    """
    Gets the content API.
    """
    return _get_client('axilent.content',library=False)

def get_content_resource():
    """
    Gets the deployed content resource.
    """
    return _get_resource('axilent.content','content',library=False)

def get_content_keys():
    """
    Gets the keys for the conten type.
    """
    cfg = get_cfg()
    api = get_content_api()
    keys = api.getcontentkeys(content_type_slug=slugify(cfg.get('Connection','content_type')))
    return keys

def get_content_item(key):
    """
    Gets deployed content for the specified key.
    """
    cfg = get_cfg()
    resource = get_content_resource()
    return resource.get(params={'content_type_slug':slugify(cfg.get('Connection','content_type')),'content_key':key})



"""
Uploads documents to Axilent.
"""
import markdown
from dox.config import get_cfg, get_keymap, get_keyfields
from dox.client import get_content_library_resource

def upload_document(path,key=None):
    """
    Uploads the specified document, returns the key of the uploaded document.
    """
    # 1. Prepare data for upload
    cfg = get_cfg()
    body_field_name = cfg.get('Connection','body_field')
    data = {}
    md = markdown.Markdown(extensions=['meta','tables'])
    with open(path) as docfile:
        body = md.convert(docfile.read())
        
        for field_name, field_value in md.Meta.items():
            data[field_name] = field_value[0]
        
        data[body_field_name] = body
    
    
    # Upload Content
    return_key = None
    resource = get_content_library_resource()
    if key:
        resource.put(data={'content':data,
                           'content_type':cfg.get('Connection','content_type'),
                           'project':cfg.get('Connection','project'),
                           'key':key})
        
        return (key,False) # no new document created
    else:
        response = resource.post(data={'content':data,
                                       'project':cfg.get('Connection','project'),
                                       'content_type':cfg.get('Connection','content_type')})
        created_content_type, created_key = response['created_content'].split(':')
        
        return (created_key,True) # new document created
    
def find_key(path,keymap,keyfields):
    """
    Finds a key for the specified document if it exists.
    """
    # First try to pull key from the keymap
    try:
        return keymap[path]
    except KeyError:
        pass
    
    # Try to find it from a keyfield
    try:
        cfg = get_cfg()
        keyfield_name = cfg.get('Connection','key_field')
        md = markdown.Markdown(extensions=['meta','tables'])
        with open(path) as docfile:
            md.convert(docfile.read())
            keyfield = md.Meta[keyfield_name][0]
            return keyfields[keyfield]
    except KeyError:
        pass
    
    # Give up
    return None

def extract_keyfield(path):
    """
    Extracts the keyfield value from the document.
    """
    cfg = get_cfg()
    keyfield_name = cfg.get('Connection','key_field')
    md = markdown.Markdown(extensions=['meta','tables'])
    with open(path) as docfile:
        md.convert(docfile.read())
        try:
            keyfield = md.Meta[keyfield_name][0]
            return keyfield
        except KeyError:
            return None



#!/usr/bin/env python
try:
    from setuptools import setup 
except ImportError, err:
    from distutils.core import setup

from dox import VERSION

setup(
    name='Axilent-Dox',
    version='.'.join(map(str,VERSION)),
    description='Markdown-oriented content authoring for Axilent.',
    packages=['dox'],
    include_package_data=True,
    license='BSD',
    author='Loren Davie',
    author_email='code@axilent.com',
    url='https://github.com/Axilent/Dox',
    install_requires=['sharrock','Markdown==2.0.1'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Topic :: Text Processing',
    ],
    scripts=['scripts/dox'],
)